package de.mpg.mpimet.yacgui.utils;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.text.SimpleDateFormat;
import java.util.Arrays;
import java.util.Date;

/**
 * Copyright (c) 2013 Max-Planck-Institute for Meteorology, Hamburg.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *   - Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 *   - Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 *   - Neither the name of Oracle or the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * Helper logger class for feeling like a master of logs. :)
 * 
 * @author Maxim Yastremsky
 */

public class Logger {
	
	private static final String DEFAULT_LOG_FILENAME = "CouplingGui_log.txt";
	public static final String DATE_FORMAT = "yyyy-MM-dd_HH:mm:ss";
	
	public enum LogLevel {
		DEBUG,
		INFO,
		WARN,
		ERROR,
		OFF;
	}
	
	public enum LogOutput {
		SYSTEM_OUT,
		FILE;
	}
	
	public static LogLevel currentLevel = LogLevel.DEBUG;
	public static LogOutput currentOutput = LogOutput.SYSTEM_OUT;
	
	private static BufferedWriter writer;
	private static File logFile;
	
	public static boolean logTime = true;
	public static boolean logLogLevel = true;
	
	public static void logDebug(String logInfo) {
		if (currentLevel.ordinal() <= LogLevel.DEBUG.ordinal()) {
			log(LogLevel.DEBUG, logInfo);
		}
	}
	
	private static void log(LogLevel level, String logInfo) {
		log(level, logInfo, null);
	}
	
	private static void log(LogLevel level, String logInfo, Exception exp) {
		
		if (logInfo == null || logInfo.equals("")) {
			return;
		}
		
		StringBuilder builder = new StringBuilder();
		if (logTime) {
			builder.append(getTimeLog() + "; ");
		}
		
		if (logLogLevel) {
			builder.append(level.toString() + ": ");
		}
		
		builder.append(logInfo);
		
		switch (currentOutput) {
		case SYSTEM_OUT:
			System.out.println(builder.toString());
			
			if (exp != null) {
				exp.printStackTrace(System.out);
			}
			break;
		case FILE:
			try {
				if (logFile == null || !logFile.exists()) { 
					logFile = new File(DEFAULT_LOG_FILENAME);
					
					if (!logFile.exists()) {
						System.out.println("creating controller log file");
					}
					writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(logFile, true), "utf-8"));
				}
				
				writer.write(builder.toString());
				
				if (exp != null) {
					writer.write(Arrays.toString(exp.getStackTrace()));
				}
				writer.newLine();
				writer.flush();
			} catch (Exception e) {
				e.printStackTrace();
			}
			break;
		default:
			break;
		}
		
	}

	public static String getTimeLog() {
		return new SimpleDateFormat(DATE_FORMAT).format(new Date());
	}

	public static void logInfo(String logInfo) {
		if (currentLevel.ordinal() <= LogLevel.INFO.ordinal()) {
			log(LogLevel.INFO, logInfo);
		}
	}
	
	public static void logWarn(String logInfo) {
		if (currentLevel.ordinal() <= LogLevel.WARN.ordinal()) {
			log(LogLevel.WARN, logInfo);
		}
	}
	
	public static void logError(String logInfo) {
		if (currentLevel.ordinal() <= LogLevel.ERROR.ordinal()) {
			log(LogLevel.ERROR, logInfo);
		}
	}
	
	public static void logError(String logInfo, Exception exc) {
		if (currentLevel.ordinal() <= LogLevel.ERROR.ordinal()) {
			log(LogLevel.ERROR, logInfo, exc);
		}
	}
}
